@@ -23,9 +23,12 @@ struct wpa_driver_wext_data {
 	int ioctl_sock;
 	int mlme_sock;
 	char ifname[IFNAMSIZ + 1];
+	char phyname[32];
 	int ifindex;
 	int ifindex2;
 	int if_removed;
+	int if_disabled;
+	struct rfkill_data *rfkill;
 	u8 *assoc_req_ies;
 	size_t assoc_req_ies_len;
 	u8 *assoc_resp_ies;
@@ -45,6 +48,8 @@ struct wpa_driver_wext_data {
 	int scan_complete_events;
 
 	int cfg80211; /* whether driver is using cfg80211 */
+
+	u8 max_level;
 };
 
 int wpa_driver_wext_get_bssid(void *priv, u8 *bssid);
